;************************************************************************************************
; Fast AVR FFT - FFT related math routines
;************************************************************************************************
; Fast FFT routines in 16-bit fixed point resolution. Cooley-Tukey DIT radix 2 algorithm.  
; All functions uses common global complex work array:
;   int16_t FFTwarr[FFTN*2]; 
; where FFTN is FFT size, memory organisation in array:
;   re{0},im{0},re{1},im{1},...,re{FFTN-1},im{FFTN-1}
; Routines should work for FFTN={32,64,128,256,512}, lower resolutions are also possible, 
; but there are no lookups generated for them.
;
; Usage:
; ------
; 1. fill im{} items of FFTwarr[] with signal to process
; 2. call function FFT_input()
;     - this function does bit reversion and applies selected window      
;     - then optionally duplicates re{} to im{} to increase amplitude by 2^0.5 (better SNR) 
; 3. call FFT_dit_r2() to do main FFT DIT radix 2 transformation
;     - complex results are stored in FFTwarr and are already in correct order
; 4. call FFT_out_lin() or FFT_out_log() to create linear or log scale amplitude spectrum from 
;    first half of complex spectrum, amplitudes "r" are stored into FFTwarr[] in this order:
;      FFTwarr[0]=r[0],FFTwarr[1]=r[1],FFTwarr[2]=r[2],... 
;
; Options (FFT.h):
; ----------------
; FFTN     - FFT size	     
; FFTW     - window to apply
; FFTdupim - duplicate input signal to both re{} and im{} part of FFTwarr[]
;          - this increases input signal amplitude by 2^0.5 so it'll have better SNR
; FFTprec  - increases algorithm precision (lower algorithm noise), needs some more CPU power
; FFTopt   - increases algorithm performance (bigger code size)
; FFTfsqrt - uses fast lookup sqrt() insted of precise one
;
; Files:
; ------
; Whole FFT library consist of several files:
; 1. FFT_routines.s - main FFT algorithm routines
; 2. FFT_macros.inc - some higher level FFT related macros
; 3. FFT_tables.inc - FFT lookups
; 4. FFT_math.s     - FFT related arithmetics
; 5. FFT.c          - global FFTwarr[] array and some demo functions
; 6. FFT.h          - common FFT library header file, algorithm parameters
; 7. macros.inc     - basic assembler macros, neccessary for all other assembler modules
;
;
;
; Measured execution times for FFTfsqrt=0, FFTprec=0, FFTdupim=1, XTAL=16MHz:
;
; FFTopt=0, linear output:
; FFTN   input  transformation  output  total   performance
; [-]    [ms]   [ms]            [ms]    [ms]    [kp/s]
; 32     0.118   0.707          0.601    1.426  22.440
; 64     0.235   1.670          1.203    3.108  20.592
; 128    0.485   3.855          2.410    6.750  18.963
; 256    0.969   8.745          4.821   14.535  17.613
; 512    1.937  19.567          9.643   31.147  16.438
;
; FFTopt=1, linear output:
; FFTN   input  transformation  output  total   performance
; [-]    [ms]   [ms]            [ms]    [ms]    [kp/s]
; 32     0.118   0.500          0.601    1.218  26.273
; 64     0.235   1.251          1.204    2.690  23.796
; 128    0.485   3.012          2.409    5.906  21.673
; 256    0.969   7.057          4.820   12.845  19.930
; 512    1.937  16.190          9.643   27.769  18.437
;
; FFTopt=0, log scale output:
; FFTN   input  transformation  output   total   performance
; [-]    [ms]   [ms]            [ms]     [ms]    [kp/s]
; 32     0.118   0.707           0.655    1.480  21.622
; 64     0.235   1.670           1.305    3.210  19.935
; 128    0.485   3.855           2.613    6.953  18.408
; 256    0.969   8.745           5.227   14.941  17.143
; 512    1.937  19.567          10.458   31.962  16.019
;
; FFTopt=1, log scale output:
; FFTN   input  transformation  output   total   performance
; [-]    [ms]   [ms]            [ms]     [ms]    [kp/s]
; 32     0.118   0.500           0.655    1.272  25.157
; 64     0.235   1.250           1.306    2.791  22.931
; 128    0.485   3.012           2.609    6.106  20.963
; 256    0.969   7.056           5.216   13.241  19.334
; 512    1.937  16.190          10.431   28.558  17.928
;
; Performance for FFTprec=1 is 12% lower or so. 
;
; (c) 2012, Stanislav Maslan, s.maslan@seznam.cz
; http://www.elektronika.kvalitne.cz/ATMEL/necoteorie/transformation/AVRFFT/AVRFFT.html
;************************************************************************************************
#include <avr/io.h>

#include "FFT.h"

// general macros
#include "macros.inc"




;================================================================================================
;	32bit square root (526..542clk)
;	R16:R17 <= sqrt(R2:R3:R4:R5)
;	Author: ChaN (downloaded from ELM web (technical notes section))
;	        http://elm-chan.org/
;
;================================================================================================
;	Routine is made for AVRFFT only, it won't work as regular GCC function! 
;	Main reason for this is that it passes parameters in different registers.
;================================================================================================
	.global	FFT_sqrt32c_asm
	.func	FFT_sqrt32c_asm
	
FFT_sqrt32c_asm:	clr	R6
	clr	R7
	clr	R8
	clr	R9
	ldi	R16,1
	ldi	R17,0
	clr	R18
	clr	R19
	ldi	R20,16
	lsl	R2
	rol	R3
	rol	R4
	rol	R5
	rol	R6
	rol	R7
	rol	R8
	rol	R9
	lsl	R2
	rol	R3
	rol	R4
	rol	R5
	rol	R6
	rol	R7
	rol	R8
	rol	R9
	brpl	PC+(6)*2
	add	R6,R16
	adc	R7,R17
	adc	R8,R18
	adc	R9,R19
	rjmp	PC+(5)*2
	sub	R6,R16
	sbc	R7,R17
	sbc	R8,R18
	sbc	R9,R19
	lsl	R16
	rol	R17
	rol	R18
	andi	R16,0b11111000
	ori	R16,0b00000101
	sbrc	R9,7
	subi	R16,2
	dec	R20
	brne	PC-(34)*2
	lsr	R18
	ror	R17
	ror	R16
	lsr	R18
	ror	R17
	ror	R16
	ret
	.endfunc
	
;------------------------------------------------------------------------------------------------
;	GCC compatible wrapper for FFT_sqrt32c_asm() routine:
;
;	uint16_t FFT_sqrt32c(uint32_t x)
;------------------------------------------------------------------------------------------------	
	.global	FFT_sqrt32c
	.func	FFT_sqrt32c
	
FFT_sqrt32c:	movw	R30,R2
	movw	R26,R4
	movw	R0,R8
	movw	R2,R22	;low(x)
	movw	R4,R24	;high(x)
	movw	R22,R16
	movw	R24,R6
	;
	rcall	FFT_sqrt32c_asm	;do main sqrt()
	movw	R6,R24
	movw	R24,R16	;16bit output
	;
	movw	R16,R22
	movw	R8,R0
	movw	R4,R26
	movw	R2,R30
	clr	R1
	ret
	.endfunc

	


;================================================================================================
;	y=sqrt32(x) [db], where:
;	 - output: y=R16:R17
;	 - input: x=R2:R3:R4:R5
;	using R15 as zero reg
;
;	Worst precision is some 0.6% 
;
;	Author: Stanislav Maslan, s.maslan@seznam.cz
;
;================================================================================================
;	Routine is made for AVRFFT only, it won't work as regular GCC function! 
;	Main reason for this is that it passes parameters in different registers.
;================================================================================================
	.global	FFT_sqrt32f_asm
	.func	FFT_sqrt32f_asm

FFT_sqrt32f_asm:	clr	R15	;0
	tst	R5
	breq	sqrt32fn32
	;=== full 32-bit ===
	mov	R2,R4
	mov	R3,R5	;x>>=16
	rcall	sqrt32int	;s=sqrt(x)>>12
	ldi	R22,16
	mul	R20,R22
	movw	R16,R0
	mul	R21,R22
	add	R17,R0	;y=s<<12
	ret
	
sqrt32fn32:	tst	R4
	breq	sqrt32fn24
	;=== 24-bit ===
	mov	R2,R3
	mov	R3,R4	;x>>=8
	rcall	sqrt32int	;s=sqrt(x)>>8
	movw	R16,R20	;y=s<<8
	ret
	
sqrt32fn24:	tst	R3
	breq	sqrt32fn16
	;=== 16-bit ===
	rcall	sqrt32int	;s=sqrt(x)>>4
	ldi	R22,16
	mul	R21,R22
	movw	R16,R0
	mul	R20,R22
	add	R16,R1
	adc	R17,R15	;y=s<<4
	ret
	;=== 8-bit ===
sqrt32fn16:	ldi	R16,8
	cp	R2,R16
	brsh	sqrt32fn16lp
	;x<8 - switch to higher precision table
	ldi	R16,16*2
	mul	R2,R16
	movw	R30,R0
	subiw	R30,-(TSqrt256p1+1)
	rjmp	sqrt32fn16hp
	;x>8 - standard precision table
sqrt32fn16lp:	ldi	R16,8
	sub	R2,R16
	ldi	R16,2
	mul	R16,R2
	movw	R30,R0
	subiw	R30,-(TSqrt256p2+1)
sqrt32fn16hp:	lpm	R16,Z
	clr	R17	;y=sqrt(x)
	ret
	

;------------------------------------------------------------------------------------------------
;	return y=sqrt(x) from lookup table where: 
;	int part: x=R3
;	fraction: f=R2 [fx 0.8]
;	output: y=R20:R21 [fx 8.8]
;	using R15 as zero reg 
;------------------------------------------------------------------------------------------------
sqrt32int:	ldi	R16,8
	cp	R3,R16
	brsh	sqrt32intLP
	;x<8 - use higher precision lookup
	ldi	R22,16
	clr	R18
	mul	R2,R22
	movw	R16,R0
	mul	R3,R22
	addw	R17,R0	;(f:x)<<4
	mov	R2,R16	;f=((f:x)<<4)&0x000FF
	mov	R30,R17
	mov	R31,R18
	lslw	R30	;x=((f:x)>>4)&0x0FFFF
	subiw	R30,-(TSqrt256p1)	;Z=TSqrt256p1[x]	
	rjmp	sqrt32intHP
	;x>=8 - use standard precision loopup 
sqrt32intLP:	ldi	R16,8
	sub	R3,R16
	ldi	R16,2
	mul	R3,R16
	movw	R30,R0
	subiw	R30,-(TSqrt256p2)	;Z=TSqrt256p2[x]
sqrt32intHP:	lpmw	R20,Z+	;y1=sqrt(x)=*Z++ [fx 8.8]
	lpmw	R22,Z+	;y2=sqrt(x+1)=*Z++ [fx 8.8]
	subw	R22,R20	;dy=y2-y1 [fx 8.8]
	
	mul	R23,R2	
	addw	R20,R0
	mul	R22,R2
	add	R20,R1
	adc	R21,R15	;y=sqrt(y1+f*dy) [fx 8.8]
	ret	

	;sqrt(x) [fx 8.8], for x=0,(1/16),(2/16)...(127/16),8 
TSqrt256p1:	.dw	0,64,91,111,128,143,157,169
	.dw	181,192,202,212,222,231,239,248
	.dw	256,264,272,279,286,293,300,307
	.dw	314,320,326,333,339,345,351,356
	.dw	362,368,373,379,384,389,395,400
	.dw	405,410,415,420,425,429,434,439
	.dw	443,448,453,457,462,466,470,475
	.dw	479,483,487,492,496,500,504,508
	.dw	512,516,520,524,528,532,535,539
	.dw	543,547,551,554,558,562,565,569
	.dw	572,576,580,583,587,590,594,597
	.dw	600,604,607,611,614,617,621,624
	.dw	627,630,634,637,640,643,646,650
	.dw	653,656,659,662,665,668,671,674
	.dw	677,680,683,686,689,692,695,698
	.dw	701,704,707,710,713,716,718,721
	.dw	724
	;sqrt(x) [fx 8.8], for x=8...256
TSqrt256p2:	.dw	724,768,810,849,887,923,958,991
	.dw	1024,1056,1086,1116,1145,1173,1201,1228
	.dw	1254,1280,1305,1330,1355,1379,1402,1425
	.dw	1448,1471,1493,1515,1536,1557,1578,1599
	.dw	1619,1639,1659,1679,1698,1717,1736,1755
	.dw	1774,1792,1810,1828,1846,1864,1881,1899
	.dw	1916,1933,1950,1966,1983,1999,2016,2032
	.dw	2048,2064,2080,2095,2111,2126,2142,2157
	.dw	2172,2187,2202,2217,2232,2246,2261,2275
	.dw	2290,2304,2318,2332,2346,2360,2374,2388
	.dw	2401,2415,2429,2442,2455,2469,2482,2495
	.dw	2508,2521,2534,2547,2560,2573,2585,2598
	.dw	2611,2623,2636,2648,2660,2673,2685,2697
	.dw	2709,2721,2733,2745,2757,2769,2781,2793
	.dw	2804,2816,2828,2839,2851,2862,2874,2885
	.dw	2896,2908,2919,2930,2941,2952,2963,2974
	.dw	2985,2996,3007,3018,3029,3040,3051,3061
	.dw	3072,3083,3093,3104,3114,3125,3135,3146
	.dw	3156,3167,3177,3187,3197,3208,3218,3228
	.dw	3238,3248,3258,3268,3278,3288,3298,3308
	.dw	3318,3328,3338,3348,3357,3367,3377,3387
	.dw	3396,3406,3415,3425,3435,3444,3454,3463
	.dw	3473,3482,3491,3501,3510,3519,3529,3538
	.dw	3547,3556,3566,3575,3584,3593,3602,3611
	.dw	3620,3629,3638,3647,3656,3665,3674,3683
	.dw	3692,3701,3710,3719,3727,3736,3745,3754
	.dw	3762,3771,3780,3788,3797,3806,3814,3823
	.dw	3831,3840,3849,3857,3866,3874,3882,3891
	.dw	3899,3908,3916,3924,3933,3941,3949,3958
	.dw	3966,3974,3982,3991,3999,4007,4015,4023
	.dw	4031,4040,4048,4056,4064,4072,4080,4088
	.dw	4096
	.endfunc
	
;------------------------------------------------------------------------------------------------
;	GCC compatible wrapper for FFT_sqrt32f_asm() routine:
;
;	uint16_t FFT_sqrt32f(uint32_t x)
;------------------------------------------------------------------------------------------------	
	.global	FFT_sqrt32f
	.func	FFT_sqrt32f
	
FFT_sqrt32f:	movw	R26,R4
	movw	R4,R24	;high(x)
	movw	R24,R2
	movw	R2,R22	;low(x)
	mov	R24,R15
	pushw	R16
	;
	rcall	FFT_sqrt32f_asm	;do main sqrt()
	movw	R2,R22
	movw	R4,R24
	mov	R15,R24
	movw	R24,R16	;16bit output
	;
	popw	R16
	clr	R1	;zero
	ret
	.endfunc
	
	
		
	
	
	

;================================================================================================
;	y=20*log10(x/log10rm) [db], where:
;	 - output: y=signed(R16:R17) [fx 8.8]
;	 - input: x=unsinged(R16:R17)
;	using R15 as zero reg
;
;	Precision is usualy +/-1, max errors are at lowest values of lookups where
;	log function is stronly nonlinear (error<0.2% except lowest amplitudes)
;	Function is primarily designed for FFT output log scaling. 
;
;	examples (for 16383 as 0dB):
;	ldiw	R16,16383
;	rcall	log10f	;R16:R17=0 (0/256=0dB)
;	ldiw	R16,164
;	rcall	log10f	;R16:R17=-10237 (-10237/256=-39.99dB)
;
;	Author: Stanislav Maslan, s.maslan@seznam.cz
;
;================================================================================================
;	Routine is made for AVRFFT only, it won't work as regular GCC function! 
;	Main reason for this is that it passes parameters in different registers.
;================================================================================================
	;theese values are linked to lookups TLog10p1,TLog10p2,TLog10p3
	;!!! can't be changed without recalculation of lookups !!!
	.if	FFTW==wHANN
	;optimized for Hann window
	.equ	log10rm,16383	;x value for y=0dB
	.equ	log10rs0,2	;table 0 step
	.equ	log10rr0,64	;table 0 max
	.equ	log10rs1,16	;table 1 step
	.equ	log10rr1,1280	;table 1 max
	.equ	log10rr2,32767	;table 2 max
	.elif	FFTW==wHAMM
	;optimized for Hamming window
	.equ	log10rm,17694	;x value for y=0dB
	.equ	log10rs0,2	;table 0 step
	.equ	log10rr0,64	;table 0 max
	.equ	log10rs1,16	;table 1 step
	.equ	log10rr1,1280	;table 1 max
	.equ	log10rr2,35388	;table 2 max
	.elif	FFTW==wBLAC
	;optimized for Blackman window
	.equ	log10rm,13762	;x value for y=0dB
	.equ	log10rs0,2	;table 0 step
	.equ	log10rr0,64	;table 0 max
	.equ	log10rs1,16	;table 1 step
	.equ	log10rr1,1280	;table 1 max
	.equ	log10rr2,27524	;table 2 max
	.elif	FFTW==wFLAC
	;optimized for Flattop window	
	.equ	log10rm,7064	;x value for y=0dB
	.equ	log10rs0,2	;table 0 step
	.equ	log10rr0,64	;table 0 max
	.equ	log10rs1,16	;table 1 step
	.equ	log10rr1,1280	;table 1 max
	.equ	log10rr2,14128	;table 2 max
	.else
	;optimized for Rectangular window
	.equ	log10rm,32767	;x value for y=0dB
	.equ	log10rs0,2	;table 0 step
	.equ	log10rr0,64	;table 0 max
	.equ	log10rs1,16	;table 1 step
	.equ	log10rr1,1280	;table 1 max
	.equ	log10rr2,65535	;table 2 max
	.endif

	
	.global	FFT_log10f_asm
	.func	FFT_log10f_asm
	
FFT_log10f_asm:	clr	R15
	cpiwh	R16,R18,log10rr0
	brsh	log10fp1N
	;=== TLog10p1 ===
	ldi	R18,256/log10rs0
	mul	R16,R18
	movw	R16,R0
	ldiw	R30,TLog10p1
	rjmp	log10fCM
		
log10fp1N:	cpiwh	R16,R18,log10rr1
	brsh	log10fp2N
	;=== TLog10p2 ===
	subiw	R16,log10rr0
	ldi	R20,256/log10rs1
	mul	R16,R20
	movw	R18,R0
	mul	R17,R20
	add	R19,R0
	movw	R16,R18
	ldiw	R30,TLog10p2
	rjmp	log10fCM
	
	;=== TLog10p3 ===
log10fp2N:	cpiwh	R16,R18,log10rr2
	brlo	PC+(1+2)*2
	ldiw	R16,log10rr2
	subiw	R16,log10rr1
	ldiw	R30,TLog10p3
	rjmp	log10fCM
		
		

;------------------------------------------------------------------------------------------------
;	y=20*log10(z) [db], where:
;	 - output: y=R16:R17 [fx 8.8]
;	 - input: z=R16:R17 [fx 8.8]
;	    - integer: x=R17
;	    - fraction: f=R16 [0.8]
;	using R15 as zero reg
;------------------------------------------------------------------------------------------------
log10fCM:	ldi	R18,2
	mul	R17,R18
	addw	R30,R0	;lookup pointer
	;get values
	lpmw	R18,Z+	;lg1=log10(x)
	lpmw	R20,Z+	;lg2=log10(x+1)
	subw	R20,R18	;ld=lg2-lg1
	;linear interpolation
	mulsu	R21,R16
	addw	R18,R0
	mul	R20,R16
	lsl	R0	;round
	adc	R18,R1
	adc	R19,R15	;lg=lg1-ld*f
	;return result
	movw	R16,R18
	ret

	
	.if	FFTW==wHANN
	;y=20*log10(x/16383) [dB] [fx 8.8], for x=0...32 [fx 15.1]
	;y range: -84.29 to -48.16dB
TLog10p1:	.dw	-23118,-20036,-18495,-17594,-16954,-16458,-16052,-15710
	.dw	-15413,-15151,-14916,-14704,-14511,-14333,-14168,-14015
	.dw	-13871,-13737,-13609,-13489,-13375,-13267,-13163,-13064
	.dw	-12970,-12879,-12792,-12708,-12627,-12549,-12474,-12401
	.dw	-12330

	;y=20*log10(x/16383) [dB] [fx 8.8], for x=4...80 [fx 12.4]
	;y range: -48.16 to -22.14dB
TLog10p2:	.dw	-12330,-11834,-11428,-11086,-10789,-10527,-10293,-10081
	.dw	 -9887, -9709, -9544, -9391, -9248, -9113, -8986, -8865
	.dw	 -8751, -8643, -8539, -8441, -8346, -8255, -8168, -8084
	.dw	 -8003, -7925, -7850, -7777, -7706, -7638, -7571, -7507
	.dw	 -7444, -7383, -7324, -7266, -7210, -7155, -7102, -7049
	.dw	 -6998, -6948, -6899, -6851, -6805, -6759, -6714, -6670
	.dw	 -6627, -6584, -6543, -6502, -6462, -6423, -6384, -6346
	.dw	 -6308, -6272, -6236, -6200, -6165, -6130, -6097, -6063
	.dw	 -6030, -5998, -5966, -5934, -5903, -5872, -5842, -5812
	.dw	 -5783, -5754, -5725, -5697, -5669

	;y=20*log10(x/16383) [dB] [fx 8.8], for x=5...127 [fx 8.8]
	;y range: -22.14 to 6.02dB
TLog10p3:	.dw	 -5669, -5263, -4921, -4624, -4362, -4128, -3916, -3722
	.dw	 -3544, -3379, -3226, -3082, -2948, -2821, -2700, -2586
	.dw	 -2478, -2374, -2275, -2181, -2090, -2003, -1919, -1838
	.dw	 -1760, -1685, -1612, -1541, -1473, -1406, -1342, -1279
	.dw	 -1218, -1159, -1101, -1045,  -990,  -936,  -884,  -833
	.dw	  -783,  -734,  -686,  -640,  -594,  -549,  -505,  -462
	.dw	  -419,  -378,  -337,  -297,  -257,  -219,  -181,  -143
	.dw	  -107,   -70,   -35,     0,    35,    69,   102,   135
	.dw	   167,   199,   231,   262,   293,   323,   353,   382
	.dw	   411,   440,   468,   496,   524,   551,   578,   605
	.dw	   631,   657,   683,   708,   733,   758,   783,   807
	.dw	   831,   855,   878,   902,   925,   948,   970,   992
	.dw	  1015,  1037,  1058,  1080,  1101,  1122,  1143,  1164
	.dw	  1184,  1204,  1225,  1244,  1264,  1284,  1303,  1323
	.dw	  1342,  1361,  1379,  1398,  1416,  1435,  1453,  1471
	.dw	  1489,  1506,  1524,  1541


	.elif	FFTW==wHAMM
	;y=20*log10(x/17694) [dB] [fx 8.8], for x=0...32 [fx 15.1]
	;y range: -84.96 to -48.83dB
TLog10p1:	.dw	-23290,-20208,-18666,-17765,-17125,-16629,-16223,-15881
	.dw	-15584,-15322,-15088,-14876,-14682,-14504,-14339,-14186
	.dw	-14043,-13908,-13781,-13660,-13546,-13438,-13334,-13236
	.dw	-13141,-13050,-12963,-12879,-12798,-12720,-12645,-12572
	.dw	-12501

	;y=20*log10(x/17694) [dB] [fx 8.8], for x=4...80 [fx 12.4]
	;y range: -48.83 to -22.81dB
TLog10p2:	.dw	-12501,-12005,-11600,-11257,-10960,-10698,-10464,-10252
	.dw	-10058, -9880, -9716, -9562, -9419, -9284, -9157, -9037
	.dw	 -8923, -8814, -8711, -8612, -8517, -8426, -8339, -8255
	.dw	 -8174, -8096, -8021, -7948, -7877, -7809, -7743, -7678
	.dw	 -7616, -7555, -7495, -7438, -7381, -7326, -7273, -7220
	.dw	 -7169, -7119, -7070, -7023, -6976, -6930, -6885, -6841
	.dw	 -6798, -6755, -6714, -6673, -6633, -6594, -6555, -6517
	.dw	 -6480, -6443, -6407, -6371, -6336, -6302, -6268, -6234
	.dw	 -6201, -6169, -6137, -6105, -6074, -6044, -6013, -5983
	.dw	 -5954, -5925, -5896, -5868, -5840

	;y=20*log10(x/17694) [dB] [fx 8.8], for x=5...138 [fx 8.8]
	;y range: -22.81 to 6.02dB
TLog10p3:	.dw	 -5840, -5435, -5092, -4795, -4533, -4299, -4087, -3893
	.dw	 -3715, -3551, -3397, -3254, -3119, -2992, -2871, -2757
	.dw	 -2649, -2545, -2447, -2352, -2261, -2174, -2090, -2009
	.dw	 -1931, -1856, -1783, -1712, -1644, -1578, -1513, -1450
	.dw	 -1389, -1330, -1272, -1216, -1161, -1108, -1055, -1004
	.dw	  -954,  -905,  -858,  -811,  -765,  -720,  -676,  -633
	.dw	  -590,  -549,  -508,  -468,  -429,  -390,  -352,  -315
	.dw	  -278,  -242,  -206,  -171,  -137,  -103,   -69,   -36
	.dw	    -4,    28,    60,    91,   122,   152,   182,   211
	.dw	   240,   269,   297,   325,   353,   380,   407,   434
	.dw	   460,   486,   512,   537,   562,   587,   612,   636
	.dw	   660,   684,   707,   731,   754,   776,   799,   821
	.dw	   843,   865,   887,   909,   930,   951,   972,   992
	.dw	  1013,  1033,  1053,  1073,  1093,  1113,  1132,  1151
	.dw	  1170,  1189,  1208,  1227,  1245,  1263,  1282,  1300
	.dw	  1317,  1335,  1353,  1370,  1388,  1405,  1422,  1439
	.dw	  1455,  1472,  1489,  1505,  1521,  1538,  1554


	.elif	FFTW==wBLAC
	;y=20*log10(x/13762) [dB] [fx 8.8], for x=0...32 [fx 15.1]
	;y range: -82.77 to -46.65dB
TLog10p1:	.dw	-22731,-19649,-18108,-17206,-16566,-16070,-15665,-15322
	.dw	-15025,-14763,-14529,-14317,-14123,-13945,-13781,-13627
	.dw	-13484,-13349,-13222,-13102,-12988,-12879,-12776,-12677
	.dw	-12582,-12491,-12404,-12320,-12239,-12161,-12086,-12013
	.dw	-11942

	;y=20*log10(x/13762) [dB] [fx 8.8], for x=4...80 [fx 12.4]
	;y range: -46.65 to -20.63dB
TLog10p2:	.dw	-11942,-11446,-11041,-10698,-10401,-10139, -9905, -9693
	.dw	 -9500, -9322, -9157, -9003, -8860, -8725, -8598, -8478
	.dw	 -8364, -8255, -8152, -8053, -7958, -7868, -7780, -7696
	.dw	 -7616, -7537, -7462, -7389, -7319, -7250, -7184, -7119
	.dw	 -7057, -6996, -6936, -6879, -6822, -6768, -6714, -6662
	.dw	 -6610, -6561, -6512, -6464, -6417, -6371, -6326, -6282
	.dw	 -6239, -6197, -6155, -6114, -6074, -6035, -5996, -5958
	.dw	 -5921, -5884, -5848, -5812, -5777, -5743, -5709, -5675
	.dw	 -5643, -5610, -5578, -5547, -5515, -5485, -5454, -5425
	.dw	 -5395, -5366, -5337, -5309, -5281

	;y=20*log10(x/13762) [dB] [fx 8.8], for x=5...107 [fx 8.8]
	;y range: -20.63 to 6.02dB
TLog10p3:	.dw	 -5281, -4876, -4533, -4236, -3974, -3740, -3528, -3334
	.dw	 -3156, -2992, -2838, -2695, -2560, -2433, -2313, -2199
	.dw	 -2090, -1987, -1888, -1793, -1702, -1615, -1531, -1450
	.dw	 -1372, -1297, -1224, -1153, -1085, -1019,  -954,  -892
	.dw	  -831,  -771,  -714,  -657,  -602,  -549,  -497,  -445
	.dw	  -395,  -347,  -299,  -252,  -206,  -161,  -117,   -74
	.dw	   -32,    10,    51,    91,   130,   169,   207,   244
	.dw	   281,   317,   353,   388,   422,   456,   490,   523
	.dw	   555,   587,   619,   650,   680,   711,   740,   770
	.dw	   799,   828,   856,   884,   912,   939,   966,   992
	.dw	  1019,  1045,  1070,  1096,  1121,  1146,  1170,  1195
	.dw	  1219,  1243,  1266,  1289,  1312,  1335,  1358,  1380
	.dw	  1402,  1424,  1446,  1467,  1489,  1510,  1531,  1551
	
	.elif	FFTW==wFLAT
	;y=20*log10(x/7064) [dB] [fx 8.8], for x=0...32 [fx 15.1]
	;y range: -76.98 to -40.86dB
TLog10p1:	.dw	-21248,-18166,-16625,-15723,-15083,-14587,-14182,-13839
	.dw	-13542,-13280,-13046,-12834,-12640,-12462,-12298,-12144
	.dw	-12001,-11866,-11739,-11619,-11505,-11396,-11293,-11194
	.dw	-11099,-11008,-10921,-10837,-10756,-10678,-10603,-10530
	.dw	-10459

	;y=20*log10(x/7064) [dB] [fx 8.8], for x=4...80 [fx 12.4]
	;y range: -40.86 to -14.84dB
TLog10p2:	.dw	-10459, -9963, -9558, -9215, -8918, -8656, -8422, -8210
	.dw	 -8017, -7839, -7674, -7520, -7377, -7242, -7115, -6995
	.dw	 -6881, -6772, -6669, -6570, -6475, -6385, -6297, -6213
	.dw	 -6133, -6055, -5979, -5906, -5836, -5767, -5701, -5636
	.dw	 -5574, -5513, -5454, -5396, -5339, -5285, -5231, -5179
	.dw	 -5128, -5078, -5029, -4981, -4934, -4888, -4843, -4799
	.dw	 -4756, -4714, -4672, -4631, -4591, -4552, -4513, -4475
	.dw	 -4438, -4401, -4365, -4329, -4294, -4260, -4226, -4193
	.dw	 -4160, -4127, -4095, -4064, -4033, -4002, -3972, -3942
	.dw	 -3912, -3883, -3855, -3826, -3798

	;y=20*log10(x/7064) [dB] [fx 8.8], for x=5...55 [fx 8.8]
	;y range: -14.84 to 6.02dB
TLog10p3:	.dw	 -3798, -3393, -3050, -2753, -2491, -2257, -2045, -1852
	.dw	 -1674, -1509, -1355, -1212, -1077,  -950,  -830,  -716
	.dw	  -607,  -504,  -405,  -310,  -219,  -132,   -48,    32
	.dw	   111,   186,   259,   329,   398,   464,   529,   591
	.dw	   652,   712,   769,   826,   881,   934,   986,  1038
	.dw	  1087,  1136,  1184,  1231,  1277,  1322,  1366,  1409
	.dw	  1451,  1493,  1534,  1574
	

	.else
	;y=20*log10(x/32767) [dB] [fx 8.8], for x=0...32 [fx 15.1]
	;y range: -90.31 to -54.19dB
TLog10p1:	.dw	-24660,-21578,-20036,-19135,-18495,-17999,-17594,-17251
	.dw	-16954,-16692,-16458,-16246,-16052,-15874,-15710,-15556
	.dw	-15413,-15278,-15151,-15031,-14916,-14808,-14705,-14606
	.dw	-14511,-14420,-14333,-14249,-14168,-14090,-14015,-13942
	.dw	-13871

	;y=20*log10(x/32767) [dB] [fx 8.8], for x=4...80 [fx 12.4]
	;y range: -54.19 to -28.16dB
TLog10p2:	.dw	-13871,-13375,-12970,-12627,-12330,-12068,-11834,-11622
	.dw	-11429,-11251,-11086,-10932,-10789,-10654,-10527,-10407
	.dw	-10293,-10184,-10081, -9982, -9887, -9796, -9709, -9625
	.dw	 -9544, -9466, -9391, -9318, -9248, -9179, -9113, -9048
	.dw	 -8986, -8925, -8865, -8808, -8751, -8696, -8643, -8591
	.dw	 -8539, -8489, -8441, -8393, -8346, -8300, -8255, -8211
	.dw	 -8168, -8126, -8084, -8043, -8003, -7964, -7925, -7887
	.dw	 -7850, -7813, -7777, -7741, -7706, -7672, -7638, -7604
	.dw	 -7571, -7539, -7507, -7475, -7444, -7414, -7383, -7354
	.dw	 -7324, -7295, -7266, -7238, -7210

	;y=20*log10(x/32767) [dB] [fx 8.8], for x=5...255 [fx 8.8]
	;y range: -28.16 to 6.02dB
TLog10p3:	.dw	 -7210, -6805, -6462, -6165, -5903, -5669, -5457, -5263
	.dw	 -5085, -4921, -4767, -4624, -4489, -4362, -4242, -4128
	.dw	 -4019, -3916, -3817, -3722, -3631, -3544, -3460, -3379
	.dw	 -3301, -3226, -3153, -3082, -3014, -2948, -2883, -2821
	.dw	 -2760, -2700, -2643, -2586, -2531, -2478, -2425, -2374
	.dw	 -2324, -2276, -2228, -2181, -2135, -2090, -2046, -2003
	.dw	 -1961, -1919, -1878, -1838, -1799, -1760, -1722, -1685
	.dw	 -1648, -1612, -1576, -1541, -1507, -1473, -1439, -1406
	.dw	 -1374, -1342, -1310, -1279, -1249, -1218, -1189, -1159
	.dw	 -1130, -1101, -1073, -1045, -1017,  -990,  -963,  -937
	.dw	  -910,  -884,  -859,  -833,  -808,  -783,  -759,  -734
	.dw	  -710,  -686,  -663,  -640,  -617,  -594,  -571,  -549
	.dw	  -527,  -505,  -483,  -462,  -440,  -419,  -398,  -378
	.dw	  -357,  -337,  -317,  -297,  -277,  -257,  -238,  -219
	.dw	  -200,  -181,  -162,  -143,  -125,  -107,   -89,   -71
	.dw	   -53,   -35,   -17,     0,    17,    35,    52,    68
	.dw	    85,   102,   118,   135,   151,   167,   183,   199
	.dw	   215,   231,   246,   262,   277,   293,   308,   323
	.dw	   338,   353,   368,   382,   397,   411,   426,   440
	.dw	   454,   468,   482,   496,   510,   524,   538,   551
	.dw	   565,   578,   591,   605,   618,   631,   644,   657
	.dw	   670,   683,   696,   708,   721,   733,   746,   758
	.dw	   770,   783,   795,   807,   819,   831,   843,   855
	.dw	   867,   878,   890,   902,   913,   925,   936,   948
	.dw	   959,   970,   981,   992,  1004,  1015,  1026,  1036
	.dw	  1047,  1058,  1069,  1080,  1090,  1101,  1111,  1122
	.dw	  1132,  1143,  1153,  1164,  1174,  1184,  1194,  1204
	.dw	  1214,  1224,  1234,  1244,  1254,  1264,  1274,  1284
	.dw	  1294,  1303,  1313,  1322,  1332,  1342,  1351,  1360
	.dw	  1370,  1379,  1389,  1398,  1407,  1416,  1425,  1435
	.dw	  1444,  1453,  1462,  1471,  1480,  1489,  1497,  1506
	.dw	  1515,  1524,  1533,  1541
	.endif	
	.endfunc
	
;------------------------------------------------------------------------------------------------
;	GCC compatible wrapper for FFT_log10f_asm() routine:
;
;	int16_t log10db(uint16_t x)
;------------------------------------------------------------------------------------------------
	.global	FFT_log10f
	.func	FFT_log10f

FFT_log10f:	movw	R26,R16
	mov	R22,R15
	;
	movw	R16,R24	;function input
	rcall	FFT_log10f_asm
	movw	R24,R16	;function return
	;
	mov	R15,R22
	movw	R16,R26
	clr	R1	;zero
	ret
	.endfunc
	

